unit module_implementations;

{

  (C) COPYRIGHT 1999 .. 2014 Blue Hell / Jan Punter

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation;

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  For all listed email addresses :

    _dot. to be substituted by a dot      '.'
    2@t2  to be substituted by an at sign '@'


  Blue Hell is a trade mark owned by

    Jan Punter
    Oogstplein 6
    7545 HP Enschede
    the Netherlands
    http://www.iaf.nl/Users/BlueHell/
    http://bluehell.electro-music.com/
    j_dot.punter2@t2iaf_dot.nl

  All rights attributed to Blue Hell are owned by Jan Punter.
}

interface

uses

  System.SysUtils, System.Math, WaveUtils, WaveStorage,

  Globals, module_defs;

type

  TARState      = ( arIdle, arA, arR, arWait);
  TAHDState     = ( ahdIdle, ahdA, ahdH, ahdD, ahdWait);
  TADSRState    = ( adsrIdle, adsrA, adsrD, adsrS, adsrR, adsrWait);
  TEnvSpeed     = ( esSlow, esMedium, esFast);
  TRWState      = ( trwIdle, trwFired);
  TTrigOScState = ( toIdle, toRunning, toWaiting, toDone);
  TSeqInState   = ( seqIdle, seqHigh);
  TSeqStepState = ( sssIdle, sssActive, sssWaiting);

  TModConstant = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer);       override;
    function    Category: string;                                                                override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModInverter = class( TMod)
  public
    function    Category: string;                                                                override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModAdder = class( TMod)
  public
    procedure   SetDefaults;                                                                     override;
    function    Category: string;                                                                override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModMultiplier = class( TMod)
  public
    procedure   SetDefaults;                                                                     override;
    function    Category: string;                                                                override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModLogic = class( TMod)
  public
    function    Category: string;                                                                override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModAnd = class( TModLogic)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
  end;

  TModNand = class( TModLogic)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
  end;

  TModOr = class( TModLogic)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
  end;

  TModNor = class( TModLogic)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
  end;

  TModXor = class( TModLogic)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
  end;

  TModXnor = class( TModLogic)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
  end;

  TModDivider = class( TMod)
  private
    FLastInput: Boolean;
    FCounter  : Integer;
    FState    : Boolean;
  public
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNot = class( TMod)
  public
    procedure   SetDefaults;                                                                     override;
    function    Category: string;                                                                override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModAmplifier = class( TMod)
  public
    function    Category: string;                                                                override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModScaler = class( TMod)
  private
    FMode : Integer;
  public
    function    Category: string;                                                                override;
    procedure   SetParamValue( const aName: string; aValue: TSignal);                             virtual;
    procedure   SetInternal   ( const aName: string; aValue: TSignal);                           override;
    function    GetInternal   ( const aName: string): TSignal;                                   override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModOsc = class( TMod)
  private
    FAccu       : TSignal;
    FPhaseAccu  : TSignal;
    FPhaseDelta : TSignal;
    FOldRes     : Boolean;
  public
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModTrigOsc = class( TModOsc)
  private
    FState   : TTrigOScState;
    FCounter : Integer;
  public
    function    Category: string;                                                                override;
    function    InputName( anIndex: Integer): string;                                            override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
  end;

  TModLfo = class( TMod)
  private
    FAccu       : TSignal;
    FPhaseAccu  : TSignal;
    FPhaseDelta : TSignal;
    FOldRes     : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer);       override;
    function    Category: string;                                                                override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModTrigLfo = class( TModLfo)
  private
    FState   : TTrigOScState;
    FCounter : Integer;
  public
    function    Category: string;                                                                override;
    function    InputName( anIndex: Integer): string;                                            override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
  end;

  TModSine = class( TModOsc)
  public
    procedure   DoTick;                                                                          override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModSineTrig = class( TModTrigOsc)
  public
    procedure   DoTick;                                                                          override;
  end;

  TModSaw = class( TModOsc)
  public
    procedure   DoTick;                                                                          override;
  end;

  TModSawTrig = class( TModTrigOsc)
  public
    procedure   DoTick;                                                                          override;
  end;

  TModTri = class( TModOsc)
  public
    procedure   DoTick;                                                                          override;
  end;

  TModTriTrig = class( TModTrigOsc)
  public
    procedure   DoTick;                                                                          override;
  end;

  TModSquare = class( TModOsc)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModSquareTrig = class( TModTrigOsc)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModSineLfo = class( TModLfo)
  public
    procedure   DoSlowTick;                                                                      override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModSineLfoTrig = class( TModTrigLfo)
  public
    procedure   DoSlowTick;                                                                      override;
  end;

  TModSawLfo = class( TModLfo)
  public
    procedure   DoSlowTick;                                                                      override;
  end;

  TModSawLfoTrig = class( TModTrigLfo)
  public
    procedure   DoSlowTick;                                                                      override;
  end;

  TModTriLfo = class( TModLfo)
  public
    procedure   DoSlowTick;                                                                      override;
  end;

  TModTriLfoTrig = class( TModTrigLfo)
  public
    procedure   DoSlowTick;                                                                      override;
  end;

  TModSquareLfo = class( TModLfo)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModSquareLfoTrig = class( TModTrigLfo)
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModRandomWalkLfo = class( TModLfo)
  private
    FState : TRWState;
    FX     : TSignal;
    FY     : TSignal;
    FZ     : TSignal;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName( anIndex: Integer): string;                                            override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoise = class( TMod)
  public
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoiseTrig = class( TMod)
  private
    FLastTrig: Boolean;
  public
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoiseLfo = class( TMod)
  public
    constructor Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer);       override;
    function    Category: string;                                                                override;
    procedure   DoSlowTick;                                                                      override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModNoiseLfoTrig = class( TMod)
  private
    FLastTrig: Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer);       override;
    function    Category: string;                                                                override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModAverage = class( TMod)
  public
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModSVF = class( TMod)
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvAR = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
  public
    procedure   SetDefaults;                                                                     override;
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvARRetrig = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
    FArmed        : Boolean;
  public
    procedure   SetDefaults;                                                                     override;
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModEnvAHD = class( TMod)
  private
    FData       : TSignalArray;
    FState      : TAHDState;
    FTime       : Integer;
    FDuration   : Integer;
    FArmed      : Boolean;
    FLastTarget : TSignal;
    FValue      : TSignal;
    FSpeed      : TEnvSpeed;
  private
    procedure   SetSpeed( aValue: TEnvSpeed);
  public
    constructor Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer);       override;
    procedure   SetDefaults;                                                                     override;
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);             override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    Speed : TEnvSpeed read FSpeed write SetSpeed;
  end;

  TModEnvADSR = class( TMod)
  private
    FData       : TSignalArray;
    FState      : TADSRState;
    FTime       : Integer;
    FDuration   : Integer;
    FArmed      : Boolean;
    FLastTarget : TSignal;
    FValue      : TSignal;
    FSpeed      : TEnvSpeed;
  private
    procedure   SetSpeed( aValue: TEnvSpeed);
  public
    constructor Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer);       override;
    procedure   SetDefaults;                                                                     override;
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    procedure   GatherData  ( const aPrefix: string; const aCallback: TDataHandler);             override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  public
    property    Speed : TEnvSpeed read FSpeed write SetSpeed;
  end;

  TModSequencer = class( TMod)
  private
    FScroll  : Integer;
    FSteps   : Integer;
    FCounter : Integer;
    FInState : TSeqInState;
    FOldRes  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer);       override;
    procedure   SetDefaults;                                                                     override;
    function    Category: string;                                                                override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModTextSequencer = class( TMod)
  private
    FCounter : Integer;
    FInState : TSeqInState;
    FOldRes  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer);       override;
    procedure   SetDefaults;                                                                     override;
    function    Category: string;                                                                override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
    procedure   SetStepValues( const aValues: TSignals);
  end;

  TModSeqStep = class( TMod)
  private
    FTime    : TSignal;
    FState   : TSeqStepState;
    FOldRes  : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer);       override;
    procedure   SetDefaults;                                                                     override;
    function    Category: string;                                                                override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModSeqClockStep = class( TMod)
  private
    FCount      : Integer;
    FState      : TSeqStepState;
    FOldRes     : Boolean;
    FOldClock   : Boolean;
    FTrigInFwd  : Boolean;
    FTrigInBwd  : Boolean;
    FDir        : Boolean;
    FTrigOutFwd : Boolean;
    FTrigOutBwd : Boolean;
  public
    constructor Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer);       override;
    procedure   SetDefaults;                                                                     override;
    function    Category: string;                                                                override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
  end;

  TModDelay = class( TMod)
  private
    FDlyTime   : TSignal;
    FMemory    : TSignals;
    FInPointer : Integer;
  public
    procedure   SetDefaults;                                                                     override;
    function    Category: string;                                                                override;
    destructor  Destroy;                                                                         override;
    procedure   SetInternal( const aName: string; aValue: TSignal);                              override;
    function    GetInternal( const aName: string): TSignal;                                      override;
    procedure   DoTick;                                                                          override;
    function    InputName   ( anIndex: Integer): string;                                         override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    InternalName( anIndex: Integer): string;                                         override;
  end;

  TModWave = class( TMod)
  private
    FFileName    : string;
    FMemory      : TSignals;
    FDuration    : string;
  private
    procedure   SetDuration( aValue: Cardinal);
    procedure   SwapMemoryFor( aMemory: TSignalArray);
    procedure   LoadFile;
    procedure   SetFileName( const aValue: string);
  public
    destructor  Destroy;                                                                         override;
    function    Category: string;                                                                override;
    procedure   SetStringValue( const aName, aValue: string);                                    override;
    function    GetStringValue( const aName: string): string;                                    override;
    procedure   GatherInfo  ( const aPrefix: string; const aCallback: TInfoHandler);             override;
    function    OutputName  ( anIndex: Integer): string;                                         override;
    function    StringName( anIndex: Integer): string;                                           override;
  public
    property    FileName: string read FFileName write SetFileName;
  end;

  TModWavePlayer = class( TModWave)
  private
    FPlayPointer : TSignal;
    FOldRes      : Boolean;
  public
    procedure   SetDefaults;                                                                     override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName( anIndex: Integer): string;                                            override;
  end;

  TModSampleAndHold = class( TMod)
  private
    FLastTrig: Boolean;
  public
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModTrackAndHold = class( TMod)
  public
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModSwitch2to1 = class( TMod)
  public
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModXFade = class( TMod)
  public
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModPan = class( TMod)
  public
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModQuantize = class( TMod)
  public
    function    Category: string;                                                                override;
    procedure   DoTick;                                                                          override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;

  TModDisplay = class( TMod)
  private
    FMax1: TSignal;
    FMax2: TSignal;
  public
    constructor Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer);       override;
    function    Category: string;                                                                override;
    procedure   DoSlowTick;                                                                      override;
    procedure   GatherLights( const aPrefix: string; const aCallback: TLightsHandler);           override;
    function    InputName ( anIndex: Integer): string;                                           override;
    function    OutputName( anIndex: Integer): string;                                           override;
  end;


  function SignalToEnvSpeed( aValue: TSignal): TEnvSpeed;



implementation



  procedure InitializeUnit;
  begin
    RegisterModuleTypes(
      [
        // Multi module
        TSynthPatch,

        // Mixers
        TModXFade,
        TModPan,

        // Signal conditioning
        TModConstant,
        TModInverter,
        TModAdder,
        TModMultiplier,
        TModAmplifier,
        TModScaler,
        TModQuantize,

        // LFOs
        TModSineLfo,
        TModSawLfo,
        TModTriLfo,
        TModSquareLfo,
        TModRandomWalkLfo,
        TModNoiseLfo,

        TModSineLfoTrig,
        TModSawLfoTrig,
        TModTriLfoTrig,
        TModSquareLfoTrig,
        TModNoiseLfoTrig,

        // Oscillators
        TModSine,
        TModSaw,
        TModTri,
        TModSquare,
        TModNoise,

        TModSineTrig,
        TModSawTrig,
        TModTriTrig,
        TModSquareTrig,
        TModNoiseTrig,

        // Logic
        TModNot,
        TModAnd,
        TModNand,
        TModOr,
        TModNor,
        TModXor,
        TModXnor,
        TModDivider,

        // Delay
        TModDelay,
        TModWavePlayer,

        // Switch
        TModSampleAndHold,
        TModTrackAndHold,
        TModSwitch2to1,

        // Envelopes
        TModEnvAR,
        TModEnvARRetrig,
        TModEnvAHD,
        TModEnvADSR,

        // Filters
        TModAverage,
        TModSVF,

        // Utility
        TModDisplay,

        // Sequencer
        TModSequencer,
        TModTextSequencer,
        TModSeqStep,
        TModSeqClockStep

      ]
    );
  end;

  procedure FinalizeUnit;
  begin
  end;

  function SignalToEnvSpeed( aValue: TSignal): TEnvSpeed;
  begin
    if AlmostEqual( aValue, 0, 0.45)
    then Result := esFast
    else if AlmostEqual( aValue, 2, 0.45)
    then Result := esSlow
    else Result := esMedium;
  end;


{ ========
  TModConstant = class( TMod)
  public
}

    constructor TModConstant.Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := False;
    end;

    function    TModConstant.Category: string; // override;
    begin
      Result := 'signal conditioning';
    end;

    procedure   TModConstant.DoSlowTick; // override;
    begin
      if SignalToLogic( FInCache[ 1])
      then FOutCache[ 0] := 0
      else FOutCache[ 0] := FInCache[ 0];
    end;

    function    TModConstant.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'value';
        1 :  Result := 'mute';
        else Result := NO_INPUT;
      end;
    end;

    function    TModConstant.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModInverter = class( TMod)
  public
}

    function    TModInverter.Category: string; // override;
    begin
      Result := 'signal conditioning';
    end;

    procedure   TModInverter.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      // todo: work
    end;

    function    TModInverter.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'inputcount')
      then Result := InputCount
      else Result := inherited;
    end;

    procedure   TModInverter.DoTick; // override;
    var
      i : Integer;
    begin
      for i := 0 to InputCount - 1
      do FOutCache[ i] := - FInCache[ i];
    end;

    function    TModInverter.InputName( anIndex: Integer): string; // override;
    begin
      if ( anIndex >= 0) and ( anIndex < InputCount)
      then Result := Format( 'in%d', [ anIndex + 1], AppLocale)
      else Result := NO_INPUT;
    end;

    function    TModInverter.OutputName( anIndex: Integer): string; // override;
    begin
      if ( anIndex >= 0) and ( anIndex < InputCount)
      then Result := Format( 'out%d', [ anIndex + 1], AppLocale)
      else Result := NO_OUTPUT;
    end;

    function    TModInverter.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'inputcount';
        else Result := NO_INTERNAL;
      end;
    end;

{ ========
  TModAdder = class( TMod)
  public
}

    procedure   TModAdder.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := 1 to InputCount - 1          // Make sure unused inputs are set to 0, so it'll 'work'
      do FInCache[ i] := 0;
    end;

    function    TModAdder.Category: string; // override;
    begin
      Result := 'signal conditioning';
    end;

    procedure   TModAdder.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      // todo: work
    end;

    function    TModAdder.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'inputcount')
      then Result := InputCount
      else Result := inherited;
    end;

    procedure   TModAdder.DoTick; // override;
    var
      i : Integer;
      R : TSignal;
    begin
      R := 0.0;
      if not SignalToLogic( FInCache[ 0])
      then begin
        for i := 1 to InputCount - 1
        do R := R + FInCache[ i];
      end;
      FOutCache[ 0] := R;
    end;

    function    TModAdder.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'mute';
        else begin
          if ( anIndex >= 1) and ( anIndex < InputCount)
          then Result := Format( 'in%d', [ anIndex], AppLocale)
          else Result := NO_INPUT;
        end;
      end;
    end;

    function    TModAdder.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;

    function    TModAdder.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'inputcount';
        else Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModMultiplier = class( TMod)
  public
}

    procedure   TModMultiplier.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := 1 to InputCount - 1          // Make sure unused inputs are set to 1, so it'll 'work'
      do FInCache[ i] := 1;
    end;

    function    TModMultiplier.Category: string; // override;
    begin
      Result := 'signal conditioning';
    end;

    procedure   TModMultiplier.SetInternal( const aName: string; aValue: TSignal); // override;
    var
      OldCount: Integer;
      NewCount: Integer;
    begin
      if SameText( aName, 'inputcount')
      then begin
        OldCount := Ceil( GetInternal( 'inputcount'));
        NewCount := Max( 1, Ceil( aValue));
        if OldCount <> NewCount
        then begin
          // do ne: need to lock the wave player here ...
          while not GrabTickLock do;
          try
            // todo: need to remove invalid wires here
            // todo: need to set the actual input count here

            // todo: need to redraw the editor module here with changed connector counts
            // todo: need to redraw the patch here
            // hmm .. but the editor sent this message .Self. and yet it is all editor stuff o_O
            // k .. the editor does not know about wave player locks ...
            // todo: need a method to actually set input count
          finally
            // do ne: need to unlock the wave player here
            ReleaseTickLock;
          end;
        end;
      end;
    end;

    function    TModMultiplier.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'inputcount')
      then Result := InputCount
      else Result := inherited;
    end;

    procedure   TModMultiplier.DoTick; // override;
    var
      i : Integer;
      R : TSignal;
    begin
      if GrabTickLock
      then begin
        try
          if SignalToLogic( FInCache[ 0])
          then R := 0.0
          else begin
            R := 1.0;
            for i := 1 to InputCount - 1
            do R := R * FInCache[ i];
          end;
          FOutCache[ 0] := AntiRunAway( R);
        finally
          ReleaseTickLock;
        end;
      end;
    end;

    function    TModMultiplier.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'mute';
        else begin
          if ( anIndex >= 1) and ( anIndex < InputCount)
          then Result := Format( 'in%d', [ anIndex], AppLocale)
          else Result := NO_INPUT;
        end;
      end;
    end;

    function    TModMultiplier.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;

    function    TModMultiplier.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'inputcount';
        else Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModLogic = class( TMod)
  public
}

    function    TModLogic.Category: string; // override;
    begin
      Result := 'logic';
    end;

    procedure   TModLogic.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      // todo: work
    end;

    function    TModLogic.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'inputcount')
      then Result := InputCount
      else Result := inherited;
    end;

    procedure   TModLogic.DoTick; // override;
    var
      i : Integer;
      R : Boolean;
    begin
      R := True;
      for i := 0 to InputCount - 1
      do R := R and SignalToLogic( FInCache[ i]);
      FOutCache[ 0] := LogicToSignal( R);
    end;

    procedure   TModLogic.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f', [ aPrefix, Name, FOutCache[ 0]], AppLocale));
    end;

    function    TModLogic.InputName( anIndex: Integer): string; // override;
    begin
      if ( anIndex >= 0) and ( anIndex < InputCount)
      then Result := Format( 'in%d', [ anIndex + 1], AppLocale)
      else Result := NO_INPUT;
    end;

    function    TModLogic.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;

    function    TModLogic.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'inputcount';
        else Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModAnd = class( TModLogic)
  public
}

    procedure   TModAnd.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := 0 to InputCount - 1          // Make sure unused inputs are set to 1, so it'll 'work'
      do FInCache[ i] := 1;
    end;

    procedure   TModAnd.DoTick; // override;
    var
      i : Integer;
      R : Boolean;
    begin
      R := True;
      for i := 0 to InputCount - 1
      do R := R and SignalToLogic( FInCache[ i]);
      FOutCache[ 0] := LogicToSignal( R);
    end;


{
  TModNand = class( TModLogic)
  public
}

    procedure   TModNand.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := 0 to InputCount - 1          // Make sure unused inputs are set to 1, so it'll 'work'
      do FInCache[ i] := 1;
    end;

    procedure   TModNand.DoTick; // override;
    var
      i : Integer;
      R : Boolean;
    begin
      R := True;
      for i := 0 to InputCount - 1
      do R := R and SignalToLogic( FInCache[ i]);
      FOutCache[ 0] := LogicToSignal( not R);
    end;


{ ========
  TModOr = class( TModLogic)
  public
}

    procedure   TModOr.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := 0 to InputCount - 1          // Make sure unused inputs are set to 0, so it'll 'work'
      do FInCache[ i] := 0;
    end;

    procedure   TModOr.DoTick; // override;
    var
      i : Integer;
      R : Boolean;
    begin
      R := False;
      for i := 0 to InputCount - 1
      do R := R or SignalToLogic( FInCache[ i]);
      FOutCache[ 0] := LogicToSignal( R);
    end;


{ ========
  TModNor = class( TModLogic)
  public
}

    procedure   TModNor.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := 0 to InputCount - 1          // Make sure unused inputs are set to 0, so it'll 'work'
      do FInCache[ i] := 0;
    end;

    procedure   TModNor.DoTick; // override;
    var
      i : Integer;
      R : Boolean;
    begin
      R := False;
      for i := 0 to InputCount - 1
      do R := R or SignalToLogic( FInCache[ i]);
      FOutCache[ 0] := LogicToSignal( not R);
    end;


{ ========
  TModXor = class( TModLogic)
  public
}

    procedure   TModXor.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := 0 to InputCount - 1          // Make sure unused inputs are set to 0, so it'll 'work'
      do FInCache[ i] := 0;
    end;

    procedure   TModXor.DoTick; // override;
    var
      OneCounts : Integer;
      i         : Integer;
    begin
      OneCounts := 0;
      for i := 0 to InputCount - 1
      do begin
        if SignalToLogic( FInCache[ i])
        then Inc( OneCounts);
      end;
      FOutCache[ 0] := LogicToSignal( OneCounts = 1);
    end;


{ ========
  TModXNor = class( TModLogic)
  public
}

    procedure   TModXnor.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := 0 to InputCount - 1          // Make sure unused inputs are set to 0, so it'll 'work'
      do FInCache[ i] := 0;
    end;

    procedure   TModXnor.DoTick; // override;
    var
      OneCounts : Integer;
      i         : Integer;
    begin
      OneCounts := 0;
      for i := 0 to InputCount - 1
      do begin
        if SignalToLogic( FInCache[ i])
        then Inc( OneCounts);
      end;
      FOutCache[ 0] := LogicToSignal( OneCounts <> 1);
    end;


{ ========
  TModDivider = class( TMod)
  private
    FLastInput: Boolean;
    FCounter  : Integer;
    FState    : Boolean;
  public
}

    function    TModDivider.Category: string; // override;
    begin
      Result := 'logic';
    end;

    procedure   TModDivider.DoTick; // override;
    const
      States : array[ Boolean] of TSignal = ( -1, 1);
    var
      Inp: Boolean;
    begin
      Inp := SignalToLogic( FInCache[ 1]);
      if FLastInput xor Inp
      then begin
        if FCounter <= 0
        then begin
          FCounter   := Max( 0, Round( FInCache[ 0] - 1));
          FState     := not FState;
          FOutCache[ 0] := States[ FState];
        end
        else Dec( FCounter);
      end;
      FLastInput := Inp;
    end;

    procedure   TModDivider.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f', [ aPrefix, Name, FOutCache[ 0]], AppLocale));
    end;

    function    TModDivider.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'count';
        1 :  Result := 'trig';
        else Result := NO_INPUT;
      end;
    end;

    function    TModDivider.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNot = class( TMod)
  public
}

    procedure   TModNot.SetDefaults; // override;
    var
      i : Integer;
    begin
      for i := 0 to InputCount - 1          //Set unused inputs to 1 - just to keep the lights off
      do FInCache[ i] := 1;
    end;

    function    TModNot.Category: string; // override;
    begin
      Result := 'logic';
    end;

    procedure   TModNot.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      // todo: work
    end;

    function    TModNot.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'inputcount')
      then Result := InputCount
      else Result := inherited;
    end;

    procedure   TModNot.DoTick; // override;
    var
      i : Integer;
    begin
      for i := 0 to InputCount - 1
      do FOutCache[ i] := LogicToSignal( not SignalToLogic( FInCache[ i]));
    end;

    procedure   TModNot.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      for i := 0 to InputCount - 1
      do aCallback( Self, Format( '%s/%s/out%d/%f', [ aPrefix, Name, i + 1, FOutCache[ i]], AppLocale));
    end;

    function    TModNot.InputName( anIndex: Integer): string; // override;
    begin
      if ( anIndex >= 0) and ( anIndex < InputCount)
      then Result := Format( 'in%d', [ anIndex + 1], AppLocale)
      else Result := NO_INPUT;
    end;

    function    TModNot.OutputName( anIndex: Integer): string; // override;
    begin
      if ( anIndex >= 0) and ( anIndex < InputCount)
      then Result := Format( 'out%d', [ anIndex + 1], AppLocale)
      else Result := NO_OUTPUT;
    end;

    function    TModNot.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'inputcount';
        else Result := NO_INTERNAL;
      end;
    end;



{ ========
  TModAmplifier = class( TMod)
  public
}

    function    TModAmplifier.Category: string; // override;
    begin
      Result := 'signal conditioning';
    end;

    procedure   TModAmplifier.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      // todo: work
    end;

    function    TModAmplifier.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'inputcount')
      then Result := InputCount - 2
      else Result := inherited;
    end;

    procedure   TModAmplifier.DoTick; // override;
    var
      i : Integer;
      G : TSignal;
    begin
      if SignalToLogic( FInCache[ 1])
      then G := 0
      else G := AntiRunAway( FInCache[ 0]);
      for i := 0 to InputCount - 3
      do FOutCache[ i] := AntiRunAway( FInCache[ i + 2] * G);
    end;

    function    TModAmplifier.InputName( anIndex: Integer): string; // override;
    begin
      if ( anIndex >= 0) and ( anIndex < InputCount)
      then begin
        case anIndex of
          0 :  Result := 'gain';
          1 :  Result := 'mute';
          else Result := Format( 'in%d', [ anIndex - 1], AppLocale);
        end;
      end
      else Result := NO_INPUT;
    end;

    function    TModAmplifier.OutputName( anIndex: Integer): string; // override;
    begin
      if ( anIndex >= 0) and ( anIndex < InputCount - 1)
      then Result := Format( 'out%d', [ anIndex + 1], AppLocale)
      else Result := NO_OUTPUT;
    end;

    function    TModAmplifier.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'inputcount';
        else Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModScaler = class( TMod)
  public
}

    function    TModScaler.Category: string; // override;
    begin
      Result := 'signal conditioning';
    end;

    procedure   TModScaler.SetParamValue( const aName: string; aValue: TSignal); // virtual;
    begin
      inherited;
      if SameText( aName, 'mode')
      then FMode := SignalToInt( aValue)
      else if FMode = 1
      then begin
      end
      else if FMode = 2
      then begin
      end;
    end;

    procedure   TModScaler.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
      // todo: work
    end;

    function    TModScaler.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'inputcount')
      then Result := InputCount - 2
      else Result := inherited;
    end;

    procedure   TModScaler.DoTick; // override;
    var
      i : Integer;
      G : TSignal;
      O : TSignal;
    begin
      if SignalToLogic( FInCache[ 3])
      then begin
        G := 0;
        O := 0;
      end
      else begin
        G := FInCache[ 0];
        O := FInCache[ 1];
      end;
      for i := 0 to InputCount - 5
      do FOutCache[ i] := AntiRunAway( G * FInCache[ i + 4] + O);
    end;

    function    TModScaler.InputName( anIndex: Integer): string; // override;
    begin
      if ( anIndex >= 0) and ( anIndex < InputCount)
      then begin
        case anIndex of
          0 :  Result := 'gain';
          1 :  Result := 'offset';
          2 :  Result := 'mode';
          3 :  Result := 'mute';
          else Result := Format( 'in%d', [ anIndex - 3], AppLocale);
        end;
      end
      else Result := NO_INPUT;
    end;

    function    TModScaler.OutputName( anIndex: Integer): string; // override;
    begin
      if ( anIndex >= 0) and ( anIndex < InputCount - 2)
      then Result := Format( 'out%d', [ anIndex + 1], AppLocale)
      else Result := NO_OUTPUT;
    end;

    function    TModScaler.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'inputcount';
        else Result := NO_INTERNAL;
      end;
    end;


{ ========
  TModOsc = class( TMod)
  private
    FAccu       : TSignal;
    FPhaseAccu  : TSignal;
    FPhaseDelta : TSignal;
    FOldRes     : Boolean;
  public
}


    function    TModOsc.Category: string; // override;
    begin
      Result := 'oscillator';
    end;

    procedure   TModOsc.DoTick; // override;
    begin
      if ( not FOldRes) and SignalToLogic( FInCache[ 3])
      then FAccu  := 0.0;
      FPhaseDelta := FrequencyToPhaseDelta( UnitsToFrequency( FInCache[ 0] + FInCache[ 1]));
      FAccu       := FModulo( FAccu + FPhaseDelta, 1.0);
      FPhaseAccu  := FModulo( FAccu + Clip( FInCache[ 2], -1, 1), 1.0);
      FOldRes     := SignalToLogic( FInCache[ 3]);
    end;

    procedure   TModOsc.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/res/%f', [ aPrefix, Name, FInCache[ 3]], AppLocale));
    end;

    function    TModOsc.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'frequency';
        1 :  Result := 'fm';
        2 :  Result := 'phase';
        3 :  Result := 'res';
        4 :  Result := 'mute';
        else Result := NO_INPUT;
      end;
    end;

    function    TModOsc.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModTrigOsc = class( TModOsc)
  private
    FState   : TTrigOscState;
    FCounter : TSignal;
  public
}

    function    TModTrigOsc.Category: string; // override;
    begin
      Result := 'oscillator triggered';
    end;

    function    TModTrigOsc.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        2 :  Result := 'trig';
        3 :  Result := 'count';
        else Result := inherited;
      end;
    end;

    procedure   TModTrigOsc.DoTick; // override;
    var
      FOldCounter : Integer;
    begin
      case FState of

        toIdle :
          begin
            FPhaseAccu := 0.0;
            FCounter   := 0;
            if SignalToLogic( FInCache[ 2])
            then FState := toRunning;
          end;

        toRunning,
        toWaiting :
          begin
            // This is almost the inherited code - except - it ignores phase

            FPhaseDelta := FrequencyToPhaseDelta( UnitsToFrequency( FInCache[ 0] + FInCache[ 1]));
            FPhaseAccu  := FModulo( FPhaseAccu + FPhaseDelta, 1.0);

            if FState = toRunning
            then begin
              FOldCounter := FCounter;
              if AlmostEqual( FPhaseAccu, 0.5, FPhaseDelta)
              then FCounter := FCounter + 1
              else if AlmostEqual( FPhaseAccu, 1.0, FPhaseDelta)
              then FCounter := FCounter + 1;
              if FCounter <> FOldCounter
              then begin
                if FCounter >= FInCache[ 3]
                then FState := toDone
                else FState := toWaiting;
              end;
            end
            else begin
              if
                not AlmostEqual( FPhaseAccu, 0.5, FPhaseDelta) and
                not AlmostEqual( FPhaseAccu, 1.0, FPhaseDelta)
              then FState := toRunning;
            end;

          end;

        toDone :
          begin
            FPhaseAccu     := 0.0;
            if FInCache[ 2] <= 0.0
            then FState := toIdle;
          end;
      end;
    end;

    procedure   TModTrigOsc.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache[ 2]], AppLocale));
    end;



{ ========
  TModLfo = class( TMod)
  private
    FAccu       : TSignal;
    FPhaseAccu  : TSignal;
    FPhaseDelta : TSignal;
    FOldRes     : Boolean;
  public
}


    constructor TModLfo.Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := False;
    end;

    function    TModLfo.Category: string; // override;
    begin
      Result := 'LFO';
    end;

    procedure   TModLfo.DoSlowTick; // override;
    begin
      if ( not FOldRes) and SignalToLogic( FInCache[ 3])
      then FAccu     := 0.0;
      FPhaseDelta    := SlowFrequencyToPhaseDelta( SlowUnitsToFrequency( FInCache[ 0] + FInCache[ 1]));
      FAccu          := FModulo( FAccu + FPhaseDelta, 1.0);
      FPhaseAccu     := FModulo( FAccu + Clip( FInCache[ 2], -1, 1), 1);
      FOldRes        := SignalToLogic( FInCache[ 3]);
    end;

    procedure   TModLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/out/%f', [ aPrefix, Name, FOutCache[ 0]], AppLocale));
      aCallback( Self, Format( '%s/%s/res/%f', [ aPrefix, Name, FInCache [ 3]], AppLocale));
    end;

    function    TModLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'frequency';
        1 :  Result := 'fm';
        2 :  Result := 'phase';
        3 :  Result := 'res';
        4 :  Result := 'mute';
        else Result := NO_INPUT;
      end;
    end;

    function    TModLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModTrigLfo = class( TModLfo)
  private
    FState   : TTrigOscState;
    FCounter : TSignal;
  public
}

    function    TModTrigLfo.Category: string; // override;
    begin
      Result := 'LFO triggered';
    end;

    function    TModTrigLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        2 :  Result := 'trig';
        3 :  Result := 'count';
        else Result := inherited;
      end;
    end;

    procedure   TModTrigLfo.DoSlowTick; // override;
    var
      FOldCounter : Integer;
    begin
      case FState of

        toIdle :
          begin
            FPhaseAccu := 0.0;
            FCounter   := 0;
            if SignalToLogic( FInCache[ 2])
            then FState := toRunning;
          end;

        toRunning,
        toWaiting :
          begin
            // This is almost the inherited code - except - it ignores phase

            FPhaseDelta := SlowFrequencyToPhaseDelta( SlowUnitsToFrequency( FInCache[ 0] + FInCache[ 1]));
            FPhaseAccu  := FModulo( FPhaseAccu + FPhaseDelta, 1.0);

            if FState = toRunning
            then begin
              FOldCounter := FCounter;
              if AlmostEqual( FPhaseAccu, 0.5, FPhaseDelta)
              then FCounter := FCounter + 1
              else if AlmostEqual( FPhaseAccu, 1.0, FPhaseDelta)
              then FCounter := FCounter + 1;
              if FCounter <> FOldCounter
              then begin
                if FCounter >= FInCache[ 3]
                then FState := toDone
                else FState := toWaiting;
              end;
            end
            else begin
              if
                not AlmostEqual( FPhaseAccu, 0.5, FPhaseDelta) and
                not AlmostEqual( FPhaseAccu, 1.0, FPhaseDelta)
              then FState := toRunning;
            end;

          end;

        toDone :
          begin
            FPhaseAccu := 0.0;
            if FInCache[ 2] <= 0.0
            then FState := toIdle;
          end;
      end;
    end;

    procedure   TModTrigLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache [ 2]], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f' , [ aPrefix, Name, FOutCache[ 0]], AppLocale));
    end;



{ ========
  TModSine = class( TModOsc)
  public
}

    procedure   TModSine.DoTick; // override;
    var
      aVal0 : Extended;
      aVal1 : Extended;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then begin
        FOutCache[ 0] := 0;
        FOutCache[ 1] := 0;
      end
      else begin
        SinCos( PI_2 * FPhaseAccu, aVal0, aVal1);
        FOutCache[ 0] := aVal0;
        FOutCache[ 1] := aVal1;
      end
    end;

    function    TModSine.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        1 :  Result := 'outcos';
        else Result := inherited
      end;
    end;

{ ========
  TModTrigSine = class( TModTrigOsc)
  public
}

    procedure   TModSineTrig.DoTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else FOutCache[ 0] := AntiRunAway( Sin( PI_2 * FPhaseAccu));
    end;


{ ========
  TModSaw = class( TModOsc)
  public
}

    procedure   TModSaw.DoTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else FOutCache[ 0] := AntiRunAway( 2 * FPhaseAccu - 1);
    end;



{ ========
  TModTrigdSaw = class( TModTrigOsc)
  public
}

    procedure   TModSawTrig.DoTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else FOutCache[ 0] := AntiRunAway( 2 * FPhaseAccu - 1);
    end;



{ ========
  TModTri = class( TModOsc)
  public
}


    procedure   TModTri.DoTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else begin
        if FPhaseAccu <= 0.5
        then FOutCache[ 0] := AntiRunAway( 4.0 * FPhaseAccu - 1)
        else FOutCache[ 0] := AntiRunAway( 4.0 * ( 1.0 - FPhaseAccu) - 1);
      end;
    end;



{ ========
  TModTrigTri = class( TModTrigOsc)
  public
}


    procedure   TModTriTrig.DoTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else begin
        if FPhaseAccu <= 0.5
        then FOutCache[ 0] := AntiRunAway( 4.0 * FPhaseAccu - 1)
        else FOutCache[ 0] := AntiRunAway( 4.0 * ( 1.0 - FPhaseAccu) - 1);
      end;
    end;



{ ========
  TModSquare = class( TModOsc)
  public
}

    procedure   TModSquare.SetDefaults;
    begin
      inherited;
      FInCache[ 5] := 0.5; // Set default PWM to 50% so it'll 'work'
    end;

    procedure   TModSquare.DoTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else begin
        if FPhaseAccu <= ( 1.0 - FInCache[ 5])
        then FOutCache[ 0] := -1.0
        else FOutCache[ 0] :=  1.0;
      end;
    end;

    function    TModSquare.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        5 :  Result := 'pwm'
        else Result := inherited;
      end;
    end;


{ ========
  TModTrigSquare = class( TModTrigOsc)
  public
}

    procedure   TModSquareTrig.SetDefaults;
    begin
      inherited;
      FInCache[ 5] := 0.5; // Set default PWM to 50% so it'll 'work'
    end;

    procedure   TModSquareTrig.DoTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else begin
        if FPhaseAccu <= ( 1.0 - FInCache[ 5])
        then FOutCache[ 0] := -1.0
        else FOutCache[ 0] :=  1.0;
      end;
    end;

    function    TModSquareTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        5 :  Result := 'pwm'
        else Result := inherited;
      end;
    end;


{ ========
  TModSineLfo = class( TModLfo)
  public
}

    procedure   TModSineLfo.DoSlowTick; // override;
    var
      aVal0 : Extended;
      aVal1 : Extended;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then begin
        FOutCache[ 0] := 0;
        FOutCache[ 1] := 0;
      end
      else begin
        SinCos( PI_2 * FPhaseAccu, aVal0, aVal1);
        FOutCache[ 0] := aVal0;
        FOutCache[ 1] := aVal1;
      end;
    end;

    function    TModSineLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        1 :  Result := 'outcos';
        else Result := inherited
      end;
    end;



{ ========
  TModTrigSineLfo = class( TModTrigLfo)
  public
}

    procedure   TModSineLfoTrig.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else FOutCache[ 0] := AntiRunAway( Sin( PI_2 * FPhaseAccu));
    end;


{ ========
  TModSawLfo = class( TModLfo)
  public
}

    procedure   TModSawLfo.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else FOutCache[ 0] := AntiRunAway( 2 * FPhaseAccu - 1);
    end;



{ ========
  TModTrigdSawLfo = class( TModTrigLfo)
  public
}

    procedure   TModSawLfoTrig.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else FOutCache[ 0] := AntiRunAway( 2 * FPhaseAccu - 1);
    end;



{ ========
  TModTriLfo = class( TModLfo)
  public
}


    procedure   TModTriLfo.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else begin
        if FPhaseAccu <= 0.5
        then FOutCache[ 0] := AntiRunAway( 4.0 * FPhaseAccu - 1)
        else FOutCache[ 0] := AntiRunAway( 4.0 * ( 1.0 - FPhaseAccu) - 1);
      end;
    end;



{ ========
  TModTrigTriLfo = class( TModTrigLfo)
  public
}


    procedure   TModTriLfoTrig.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else begin
        if FPhaseAccu <= 0.5
        then FOutCache[ 0] := AntiRunAway( 4.0 * FPhaseAccu - 1)
        else FOutCache[ 0] := AntiRunAway( 4.0 * ( 1.0 - FPhaseAccu) - 1);
      end;
    end;



{ ========
  TModSquareLfo = class( TModLfo)
  public
}

    procedure   TModSquareLfo.SetDefaults;
    begin
      inherited;
      FInCache[ 5] := 0.5; // Set default PWM to 50% so it'll 'work'
    end;

    procedure   TModSquareLfo.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else begin
        if FPhaseAccu <= ( 1.0 - FInCache[ 5])
        then FOutCache[ 0] := -1.0
        else FOutCache[ 0] :=  1.0;
      end;
    end;

    function    TModSquareLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        5 :  Result := 'pwm'
        else Result := inherited;
      end;
    end;


{ ========
  TModTrigSquareLfo = class( TModTrigLfo)
  public
}

    procedure   TModSquareLfoTrig.SetDefaults;
    begin
      inherited;
      FInCache[ 5] := 0.5; // Set default PWM to 50% so it'll 'work'
    end;

    procedure   TModSquareLfoTrig.DoSlowTick; // override;
    begin
      inherited;                                // For phase calculations
      if SignalToLogic( FInCache[ 4])
      then FOutCache[ 0] := 0
      else begin
        if FPhaseAccu <= ( 1.0 - FInCache[ 5])
        then FOutCache[ 0] := -1.0
        else FOutCache[ 0] :=  1.0;
      end;
    end;

    function    TModSquareLfoTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        5 :  Result := 'pwm'
        else Result := inherited;
      end;
    end;

{ ========
  TModRandomWalkLfo = class( TModLfo)
  private
    FState : TRWState;
    FX     : TSignal;
    FY     : TSignal;
    FZ     : TSignal;
  public
}

    procedure   TModRandomWalkLfo.SetDefaults; // override;
    begin
      inherited;
    end;

    procedure   TModRandomWalkLfo.DoSlowTick; // override;
    var
      Distance : TSignal;
    begin

      FPhaseDelta := SlowFrequencyToPhaseDelta( SlowUnitsToFrequency( FInCache[ 0] + FInCache[ 1]));
      FPhaseAccu  := FModulo( FPhaseAccu + FPhaseDelta, 1.0);

      case FState of
        trwIdle :
          begin
            if FPhaseAccu >= 0.5
            then begin
              FState   := trwFired;
              if SignalToLogic( FInCache[ 3])
              then begin
                FOutCache[ 0] := 0;
                FOutCache[ 1] := 0;
                FOutCache[ 2] := 0;
              end
              else begin
                Distance := FInCache[ 2];
                case Random( 6) of
                  0 : FX := FX + Distance;
                  1 : FX := FX - Distance;
                  2 : FY := FY + Distance;
                  3 : FY := FY - Distance;
                  4 : FZ := FZ + Distance;
                  5 : FZ := FZ - Distance;
                end;
                if FX < -1 then FX := - 2 - FX else if FX > 1 then FX := 2 - FX;
                if FY < -1 then FY := - 2 - FY else if FY > 1 then FY := 2 - FY;
                if FZ < -1 then FZ := - 2 - FZ else if FZ > 1 then FZ := 2 - FZ;
                if SignalToLogic( FInCache[ 3])
                then begin
                  FOutCache[ 0] := 0;
                  FOutCache[ 1] := 0;
                  FOutCache[ 2] := 0;
                end
                else begin
                  FOutCache[ 0] := FX;
                  FOutCache[ 1] := FY;
                  FOutCache[ 2] := FZ;
                end;
              end;
            end;
          end;

        trwFired :
          begin
            if FPhaseAccu < 0.5
            then FState := trwIdle;
          end;
      end;
    end;

    procedure   TModRandomWalkLfo.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/outx/%f', [ aPrefix, Name, FOutCache[ 0]], AppLocale));
      aCallback( Self, Format( '%s/%s/outy/%f', [ aPrefix, Name, FOutCache[ 1]], AppLocale));
      aCallback( Self, Format( '%s/%s/outz/%f', [ aPrefix, Name, FOutCache[ 2]], AppLocale));
    end;

    function    TModRandomWalkLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'frequency';
        1 :  Result := 'fm';
        2 :  Result := 'distance';
        3 :  Result := 'mute';
        else Result := NO_INPUT;
      end;
    end;

    function    TModRandomWalkLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'outx';
        1 :  Result := 'outy';
        2 :  Result := 'outz';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoise = class( TMod)
  public
}

    function    TModNoise.Category: string; // override;
    begin
      Result := 'oscillator';
    end;

    procedure   TModNoise.DoTick; // override;
    var
      Color : TSignal;
    begin
      Color := FInCache[ 0];
      if Color < -0.99
      then Color := -0.99;
      if SignalToLogic( FInCache[ 1])
      then FOutCache[ 0] := 0
      else FOutCache[ 0] := AntiRunAway(( 1 - Color) * FOutCache[ 0] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color));
    end;

    function    TModNoise.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'color';
        1 :  Result := 'mute';
        else Result := NO_INPUT;
      end;
    end;

    function    TModNoise.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModTrigNoise = class( TMod)
  private
    FLastInput: TSignal;
  public
}

    function    TModNoiseTrig.Category: string; // override;
    begin
      Result := 'oscillator triggered';
    end;

    procedure   TModNoiseTrig.DoTick; // override;
    var
      Color : TSignal;
      Trig  : Boolean;
    begin
      Trig  := SignalToLogic( FInCache[ 1]);
      if not FLastTrig and Trig
      then begin
        Color := FInCache[ 0];
        if Color < -0.99
        then Color := -0.99;
        if SignalToLogic( FInCache[ 2])
        then FOutCache[ 0] := 0
        else FOutCache[ 0] := AntiRunAway(( 1 - Color) * FOutCache[ 0] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color));
        FLastTrig := Trig;
      end
      else if FLastTrig and not Trig
      then FLastTrig := Trig;
    end;

    procedure   TModNoiseTrig.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache [ 1]], AppLocale));
    end;

    function    TModNoiseTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'color';
        1 :  Result := 'trig';
        2 :  Result := 'mute';
        else Result := NO_INPUT;
      end;
    end;

    function    TModNoiseTrig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModNoiseLfo = class( TMod)
  public
}

    constructor TModNoiseLfo.Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := False;
    end;

    function    TModNoiseLfo.Category: string; // override;
    begin
      Result := 'LFO triggered';
    end;

    procedure   TModNoiseLfo.DoSlowTick; // override;
    var
      Color : TSignal;
    begin
      Color := FInCache[ 0];
      if Color < -0.99
      then Color := -0.99;
      if SignalToLogic( FInCache[ 1])
      then FOutCache[ 0] := 0
      else FOutCache[ 0] := AntiRunAway(( 1 - Color) * FOutCache[ 0] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color));
    end;

    function    TModNoiseLfo.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'color';
        1 :  Result := 'mute';
        else Result := NO_INPUT;
      end;
    end;

    function    TModNoiseLfo.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModTrigNoiseLfo = class( TMod)
  private
    FLastInput: TSignal;
  public
}

    constructor TModNoiseLfoTrig.Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := False;
    end;

    function    TModNoiseLfoTrig.Category: string; // override;
    begin
      Result := 'LFO';
    end;

    procedure   TModNoiseLfoTrig.DoSlowTick; // override;
    var
      Color : TSignal;
      Trig  : Boolean;
    begin
      Trig  := SignalToLogic( FInCache[ 1]);
      if not FLastTrig and Trig
      then begin
        Color := FInCache[ 0];
        if Color < -0.99
        then Color := -0.99;
        if SignalToLogic( FInCache[ 2])
        then FOutCache[ 0] := 0
        else FOutCache[ 0] := AntiRunAway( 1 - Color) * FOutCache[ 0] + Color * ( 1.0 - 2 * Random) / Sqrt( 1 + Color);
        FLastTrig := Trig;
      end
      else if FLastTrig and not Trig
      then FLastTrig := Trig;
    end;

    procedure   TModNoiseLfoTrig.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache [ 1]], AppLocale));
      aCallback( Self, Format( '%s/%s/out/%f' , [ aPrefix, Name, FOutCache[ 0]], AppLocale));
    end;

    function    TModNoiseLfoTrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'color';
        1 :  Result := 'trig';
        2 :  Result := 'mute';
        else Result := NO_INPUT;
      end;
    end;

    function    TModNoiseLfoTrig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModAvg = class( TMod)
  public
}

    function    TModAverage.Category: string; // override;
    begin
      Result := 'filter';
    end;

    procedure   TModAverage.DoTick; // override;
    var
      Alpha : TSignal;
      Inp   : TSignal;
      Outp  : TSignal;
    begin
      Alpha := AntiRunAway( LogAttack( FInCache[ 1], 1, 0, 1));
      Inp   := AntiRunAway( FInCache[ 0]);
      Outp  := AntiRunAway( Alpha * FOutCache[ 0] + ( 1 - Alpha) * Inp);
      FOutCache[ 0] := Outp;
      FOutCache[ 1] := Inp - Outp;
    end;

    function    TModAverage.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'in';
        1 :  Result := 'alpha';
        else Result := NO_INPUT;
      end;
    end;

    function    TModAverage.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'lp';
        1 :  Result := 'hp';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModSVF = class( TMod)
  public
}

    function    TModSVF.Category: string; // override;
    begin
      Result := 'filter';
    end;

    (*
      from    : http://www.musicdsp.org/archive.php?classid=3#92
      also see: http://www.music.mcgill.ca/~ich/classes/FiltersChap2.pdf

      input  = input buffer;
      output = output buffer;
      fs     = sampling frequency;
      fc     = cutoff frequency normally something like:
               440.0*pow(2.0, (midi_note - 69.0)/12.0);                 :: We use signals rather, Frequency = Signal * SampleRate / 2
      res    = resonance 0 to 1;
      drive  = internal distortion 0 to 0.1
      freq   = 2.0*sin(PI*MIN(0.25, fc/(fs*2)));  // the fs*2 is because it's double sampled
      damp   = MIN(2.0*(1.0 - pow(res, 0.25)), MIN(2.0, 2.0/freq - freq*0.5));
      notch  = notch output
      low    = low pass output
      high   = high pass output
      band   = band pass output
      peak   = peaking output = low - high
      --
      double sampled svf loop:
      for (i=0; i<numSamples; i++)
      {
        in    = input[i];
        notch = in - damp*band;
        low   = low + freq*band;
        high  = notch - low;
        band  = freq*high + band - drive*band*band*band;
        out   = 0.5*(notch or low or high or band or peak);
        notch = in - damp*band;
        low   = low + freq*band;
        high  = notch - low;
        band  = freq*high + band - drive*band*band*band;
        out  += 0.5*(same out as above);
        output[i] = out;
      }
    *)


    procedure   TModSVF.DoTick; // override;
    const
      FOVER = 2; // over sampling rate.
    var
      HP    : TSignal;
      BP    : TSignal;
      LP    : TSignal;
      BR    : TSignal;
      OutHP : TSignal;
      OutBP : TSignal;
      OutLP : TSignal;
      OutBR : TSignal;
      Fc    : TSignal;
      Freq  : TSignal;
      Damp  : TSignal;
      Drive : TSignal;
      i     : Integer;
    begin
      Fc   := UnitsToFrequency( FInCache[ 1] + FInCache[ 4]);
      Freq := 2 * Sin( PI * Min( 0.25, ( Fc / SYSTEM_RATE_2)));

      // damp   = MIN( 2.0 * ( 1.0 - pow( res, 0.25)), MIN( 2.0, 2.0 / freq - freq * 0.5));

      if Freq <= 0.0001
      then Damp := Min( 2.0 * ( 1.0 - Power( FInCache[ 2], 0.25)), 2.0)
      else Damp := Min( 2.0 * ( 1.0 - Power( FInCache[ 2], 0.25)), Min( 2.0, 2.0 / Freq - Freq * 0.5));

      Drive := Clip( FInCache[ 3] / 10, 0.0, 0.1);
      LP    := FOutCache[ 0];
      OutHP := 0.0;
      OutBP := 0.0;
      OutLP := 0.0;
      OutBR := 0.0;
      BP    := FOutCache[ 1];

      for i := 1 to FOVER
      do begin
        BR    := FInCache[ 0] - Damp * BP;
        LP    := LP + Freq * BP;
        HP    := BR - LP;
        BP    := Freq * HP + BP - Drive * BP * BP * BP; // with 0 <= Drive <= 0.1
        OutHP := OutHP + HP;
        OutBP := OutBP + BP;
        OutLP := OutLP + LP;
        OutBR := OutBR + BR;
      end;

      FOutCache[ 0] := AntiRunAway( OutLP / FOVER);
      FOutCache[ 1] := AntiRunAway( OutBP / FOVER);
      FOutCache[ 2] := AntiRunAway( OutHP / FOVER);
      FOutCache[ 3] := AntiRunAway( OutBR / FOVER);
    end;

    function    TModSVF.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'in';
        1 :  Result := 'frequency';
        2 :  Result := 'q';
        3 :  Result := 'distortion';
        4 :  Result := 'fm';
        else Result := NO_INPUT;
      end;
    end;

    function    TModSVF.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'lp';
        1 :  Result := 'bp';
        2 :  Result := 'hp';
        3 :  Result := 'br';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEnvAR = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
  public
}

    procedure   TModEnvAR.SetDefaults; // override;
    begin
      inherited;
      FInCache[ 3] := 1; // Set default gain to 1, so it'll work unconnected
    end;

    function    TModEnvAR.Category: string; // override;
    begin
      Result := 'envelope';
    end;

    procedure   TModEnvAR.DoTick; // override;
    var
      R : TSignal;
    begin
      case FState of

        arIdle :

          if SignalToLogic( FInCache[ 0])
          then begin
            if FInCache[ 1] <= 0
            then FCoeffAttack := 1
            else FCoeffAttack := 1 / TimeToSampleCount( FInCache[ 1]); // Linear attack
            if FInCache[ 2] <= 0
            then FCoeffRelease := -1
            else FCoeffRelease := ( Ln( 0.001) - Ln( 1.000)) / TimeToSampleCount( FInCache[ 2]); // Exp release
            // if FCoeffRelease > 0.999
            // then FCoeffRelease := 0.999;
            FState := arA;
          end;

        arA :
          begin
            R := FValue;
            if R < 0.001
            then R := 0.001;
            R := R + FCoeffAttack;
            if R > 0.999
            then begin
              FValue := 1.0;
              FState := arR;
            end
            else FValue := R;
            case SignalToInt( FInCache[ 5]) of
              1  : FOutCache[ 0] := 0;
              2  : FOutCache[ 0] := 1;
              else FOutCache[ 0] := AntiRunAway( FValue * FInCache[ 3]);
            end;
          end;

        arR :
          begin
            R := FValue;
            R := R * ( 1 + FCoeffRelease);
            if R < 0.001
            then begin
              FValue := 0.0;
              FState := arWait;
            end
            else FValue := R;
            case SignalToInt( FInCache[ 5]) of
              1  : FOutCache[ 0] := 0;
              2  : FOutCache[ 0] := 1;
              else FOutCache[ 0] := AntiRunAway( FValue * FInCache[ 3]);
            end;
          end;

        arWait :
          if not SignalToLogic( FInCache[ 0])
          then FState := arIdle;

      end;
      case SignalToInt( FInCache[ 5]) of
        1  : FOutCache[ 1] := 0;
        2  : FOutCache[ 1] := AntiRunAway( FInCache[ 4]);
        else FOutCache[ 1] := AntiRunAway( FInCache[ 4] * FOutCache[ 0]);
      end;
    end;

    procedure   TModEnvAR.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache [ 0]], AppLocale));
    end;

    function    TModEnvAR.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'trig';
        1 :  Result := 'attack';
        2 :  Result := 'release';
        3 :  Result := 'gain';
        4 :  Result := 'sigin';
        5 :  Result := 'mute';
        else Result := NO_INPUT;
      end;
    end;

    function    TModEnvAR.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        1 :  Result := 'sigout';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEnvARRetrig = class( TMod)
  private
    FState        : TARState;
    FValue        : TSignal;
    FCoeffAttack  : TSignal;
    FCoeffRelease : TSignal;
    FArmed        : Boolean;
  public
}

    procedure   TModEnvARRetrig.SetDefaults; // override;
    begin
      inherited;
      FInCache[ 3] := 1; // Set default gain to 1, so it'll work unconnected
    end;

    function    TModEnvARRetrig.Category: string; // override;
    begin
      Result := 'envelope';
    end;

    procedure   TModEnvARRetrig.DoTick; // override;
    var
      R : TSignal;
    begin
      if FArmed and SignalToLogic( FInCache[ 0])
      then FState := arIdle;

      case FState of

        arIdle :

          if SignalToLogic( FInCache[ 0])
          then begin
            FCoeffAttack  := 1 / TimeToSampleCount( FInCache[ 1]); // Linear attack
            FCoeffRelease := ( Ln( 0.001) - Ln( 1.000)) / TimeToSampleCount( FInCache[ 2]); // Exp release
            FState        := arA;
            FArmed        := False;
          end;

        arA :
          begin
            if not FArmed
            then FArmed := not SignalToLogic( FInCache[ 0]);
            R := FValue;
            if R < 0.001
            then R := 0.001;
            R := R + FCoeffAttack;
            if R > 0.999
            then begin
              FValue := 1.0;
              FState := arR;
            end
            else FValue := R;
            case SignalToInt( FInCache[ 5]) of
              1 :  FOutCache[ 0] := 0;
              2 :  FOutCache[ 0] := 1;
              else FOutCache[ 0] := AntiRunAway( FValue * FInCache[ 3]);
            end;
          end;

        arR :
          begin
            if not FArmed
            then FArmed := not SignalToLogic( FInCache[ 0]);
            R := FValue;
            R := R * ( 1 + FCoeffRelease);
            if R < 0.001
            then begin
              FValue := 0.0;
              FState := arWait;
            end
            else FValue := R;
            case SignalToInt( FInCache[ 5]) of
              1 :  FOutCache[ 0] := 0;
              2 :  FOutCache[ 0] := 1;
              else FOutCache[ 0] := AntiRunAway( FValue * FInCache[ 3]);
            end;
          end;

        arWait :
          if not SignalToLogic( FInCache[ 0])
          then FState := arIdle;

      end;
      case SignalToInt( FInCache[ 5]) of
        1 :  FOutCache[ 1] := 0;
        2 :  FOutCache[ 1] := AntiRunAway( FInCache[ 4]);
        else FOutCache[ 1] := AntiRunAway( FInCache[ 4] * FOutCache[ 0]);
      end;
    end;

    procedure   TModEnvARRetrig.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache [ 0]], AppLocale));
    end;

    function    TModEnvARRetrig.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'trig';
        1 :  Result := 'attack';
        2 :  Result := 'release';
        3 :  Result := 'gain';
        4 :  Result := 'sigin';
        5 :  Result := 'mute';
        else Result := NO_INPUT;
      end;
    end;

    function    TModEnvARRetrig.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        1 :  Result := 'sigout';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModEnvAHD = class( TMod)
  private
    FData       : TSignalArray;
    FState      : TAHDState;
    FTime       : Integer;
    FDuration   : Integer;
    FArmed      : Boolean;
    FLastTarget : TSignal;
    FValue      : TSignal;
  private
}

    procedure   TModEnvAHD.SetSpeed( aValue: TEnvSpeed);
    begin
      FSpeed := aValue;
    end;

//  public

    constructor TModEnvAHD.Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer); // override;
    begin
      inherited;
      SetLength( FData, 67);       // Should match the viewer width in the designer
      Speed := esMedium;
    end;

    procedure   TModEnvAHD.SetDefaults; // override;
    begin
      inherited;
      FInCache[ 7] := 1; // Set default gain to 1, so it'll work unconnected
    end;

    function    TModEnvAHD.Category: string; // override;
    begin
      Result := 'envelope';
    end;

    procedure   TModEnvAHD.DoTick; // override;
    var
      R      : TSignal;
      Trig   : Boolean;
      Mode   : Integer; // Gate / Trig
      AShape : Integer; // Log / Lin / Exp
      DShape : Integer; // Log / Lin / Exp
      Gain   : TSignal;
    begin
      Trig    := SignalToLogic( FInCache[ 0]);
      Mode    := SignalToInt  ( FInCache[ 1]);
      AShape  := SignalToInt  ( FInCache[ 2]);
      DShape  := SignalToInt  ( FInCache[ 3]);
      Gain    := FInCache[ 7];
      R       := FValue;

      if FArmed and Trig and ( Mode = 1)
      then FState := ahdIdle;

      case FState of

        ahdIdle :

          if Trig
          then begin
            FState      := ahdA;
            FArmed      := False;
            FLastTarget := FValue;
            FTime       := 0;
            FDuration := Round( TimeToSampleCount( FInCache[ 4]));
          end;

        ahdA :
          begin
            Inc( FTime);
            if not FArmed
            then FArmed := not Trig;
            case DShape of
              0  : R := LogAttack( FTime, FDuration, FLastTarget, 1);
              1  : R := LinAttack( FTime, FDuration, FLastTarget, 1);
              else R := ExpAttack( FTime, FDuration, FLastTarget, 1);
            end;
            if FTime >= FDuration
            then begin
              FLastTarget := R;
              FTime       := 0;
              FDuration   := Round( TimeToSampleCount( FInCache[ 5]));
              FState      := ahdH;
            end;
          end;

        ahdH :
          begin
            Inc( FTime);
            if not FArmed
            then FArmed := not Trig;
            if FTime >= FDuration
            then begin
              FLastTarget := R;
              FTime       := 0;
              FDuration   := Round( TimeToSampleCount( FInCache[ 6]));
              FState      := ahdD;
            end;
          end;

        ahdD :
          begin
            Inc( FTime);
            if not FArmed
            then FArmed := not Trig;
            case AShape of
              0  : R := LogDecay( FTime, FDuration, FLastTarget, 0);
              1  : R := LinDecay( FTime, FDuration, FLastTarget, 0);
              else R := ExpDecay( FTime, FDuration, FLastTarget, 0);
            end;
            if FTime >= FDuration
            then begin
              FLastTarget := R;
              FState      := ahdWait;
            end;
          end;

        ahdWait :
          if not Trig or FArmed
          then FState := ahdIdle;

      end;
      FValue        := R;
      case SignalToInt( FInCache[ 9]) of
        1 : begin
          FOutCache[ 0] := 0;
          FOutCache[ 1] := 0;
        end;
        2 : begin
          FOutCache[ 0] := 1;
          FOutCache[ 1] := AntiRunAway( FInCache[ 8]);
        end;
        else begin
          FOutCache[ 0] := AntiRunAway( FValue * Gain);
          FOutCache[ 1] := AntiRunAway( FInCache[ 8] * FValue * Gain); // sigout := sigin * out
        end;
      end;
    end;

    procedure   TModEnvAHD.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache [ 0]], AppLocale));
    end;

    procedure   TModEnvAHD.GatherData( const aPrefix: string; const aCallback: TDataHandler); // override;
    var
      ATime  : TSignal;
      HTime  : TSignal;
      DTime  : TSignal;
      TTime  : TSignal;
      i      : Integer;
      p      : Integer;
      AShape : Integer;
      DShape : Integer;
      R      : TSignal;
    begin
      ATime  := FInCache[ 4];
      HTime  := FInCache[ 5];
      DTime  := FInCache[ 6];
      TTime  := ATime + HTime + DTime;
      ATime  := ATime * 67 / TTime;
      HTime  := HTime * 67 / TTime;
      DTime  := DTime * 67 / TTime;
      AShape := SignalToInt  ( FInCache[ 2]);
      DShape := SignalToInt  ( FInCache[ 3]);

      i := 0;
      while i < ATime
      do begin
        case AShape of
          0  : R := LogAttack( i, ATime, 0, 1);
          1  : R := LinAttack( i, ATime, 0, 1);
          else R := ExpAttack( i, ATime, 0, 1);
        end;
        FData[ i] := R;
        Inc( i);
      end;
      while i < ATime + HTime
      do begin
        FData[ i] := 1;
        Inc( i);
      end;
      p := 0;
      while i < 67
      do begin
        case DShape of
          0  : R := LogDecay( p, DTime, 1, 0);
          1  : R := LinDecay( p, DTime, 1, 0);
          else R := ExpDecay( p, DTime, 1, 0);
        end;
        FData[ i] := R;
        Inc( i);
        Inc( p);
      end;

      aCallback( Self, Format( '%s/%s/viewer', [ aPrefix, Name], AppLocale), FData);
    end;

    function    TModEnvAHD.InputName ( anIndex: Integer): string; // override;
    begin
      case anIndex of
         0 : Result := 'trig';
         1 : Result := 'mode';
         2 : Result := 'ashape';
         3 : Result := 'dshape';
         4 : Result := 'attack';
         5 : Result := 'hold';
         6 : Result := 'decay';
         7 : Result := 'gain';
         8 : Result := 'sigin';
         9 : Result := 'mute';
        10 : Result := 'speed';
        else Result := NO_INPUT;
      end;
    end;

    function    TModEnvAHD.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        1 :  Result := 'sigout';
        else Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModEnvADSR = class( TMod)
  private
    FData       : TSignalArray;
    FState      : TADSRState;
    FTime       : Integer;
    FDuration   : Integer;
    FArmed      : Boolean;
    FLastTarget : TSignal;
    FValue      : TSignal;
  private
}

    procedure   TModEnvADSR.SetSpeed( aValue: TEnvSpeed);
    begin
      FSpeed := aValue;
    end;

//  public

    constructor TModEnvADSR.Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer); // override;
    begin
      inherited;
      SetLength( FData, 67);       // Should match the viewer width in the designer
      Speed := esMedium;
    end;

    procedure   TModEnvADSR.SetDefaults; // override;
    begin
      inherited;
      FInCache[ 9] := 1; // Set default gain to 1, so it'll work unconnected
    end;

    function    TModEnvADSR.Category: string; // override;
    begin
      Result := 'envelope';
    end;

    procedure   TModEnvADSR.DoTick; // override;
    var
      R      : TSignal;
      Trig   : Boolean;
      Mode   : Integer; // Gate / Trig
      AShape : Integer; // Log / Lin / Exp
      DShape : Integer; // Log / Lin / Exp
      RShape : Integer; // Log / Lin / Exp
      SLevel : TSignal; // 0 .. 1
      Gain   : TSignal;
    begin
      Trig    := SignalToLogic( FInCache[ 0]);
      Mode    := SignalToInt  ( FInCache[ 1]);
      AShape  := SignalToInt  ( FInCache[ 2]);
      DShape  := SignalToInt  ( FInCache[ 3]);
      RShape  := SignalToInt  ( FInCache[ 4]);
      SLevel  := FInCache[ 7];
      Gain    := FInCache[ 9];
      R       := FValue;

      if FArmed and Trig and ( Mode = 1)
      then FState := adsrIdle;

      case FState of

        adsrIdle :

          if Trig
          then begin
            FState      := adsrA;
            FArmed      := False;
            FLastTarget := FValue;
            FTime       := 0;
            FDuration   := Round( TimeToSampleCount( FInCache[ 5]));
          end;

        adsrA :
          begin
            Inc( FTime);
            if not FArmed
            then FArmed := not Trig;
            case AShape of
              0  : R := LogAttack( FTime, FDuration, FLastTarget, 1);
              1  : R := LinAttack( FTime, FDuration, FLastTarget, 1);
              else R := ExpAttack( FTime, FDuration, FLastTarget, 1);
            end;
            if FTime >= FDuration
            then begin
              FLastTarget    := R;
              FTime     := 0;
              FDuration := Round( TimeToSampleCount( FInCache[ 6]));
              FState    := adsrD;
            end;
          end;

        adsrD :
          begin
            Inc( FTime);
            if not FArmed
            then FArmed := not Trig;
            case DShape of
              0  : R := LogDecay( FTime, FDuration, FLastTarget, SLevel);
              1  : R := LinDecay( FTime, FDuration, FLastTarget, SLevel);
              else R := ExpDecay( FTime, FDuration, FLastTarget, SLevel);
            end;
            if FTime >= FDuration
            then begin
              FLastTarget := R;
              FState := adsrS;
            end;
          end;

        adsrS :
          begin
            if not FArmed
            then FArmed := not Trig;
            if FArmed
            then begin
              FLastTarget    := SLevel;
              FTime     := 0;
              FDuration := Round( TimeToSampleCount( FInCache[ 8]));
              FState    := adsrR;
            end;
          end;

        adsrR :
          begin
            Inc( FTime);
            if not FArmed
            then FArmed := not Trig;
            case RShape of
              0  : R := LogDecay( FTime, FDuration, FLastTarget, 0);
              1  : R := LinDecay( FTime, FDuration, FLastTarget, 0);
              else R := ExpDecay( FTime, FDuration, FLastTarget, 0);
            end;
            if FTime >= FDuration
            then begin
              FLastTarget := R;
              FState := adsrWait;
            end;
          end;

        adsrWait :
          if not Trig or FArmed
          then FState := adsrIdle;

      end;
      FValue := R;
      case SignalToInt( FInCache[ 11]) of
        1 : begin
          FOutCache[ 0] := 0;
          FOutCache[ 1] := 0;
        end;
        2 : begin
          FOutCache[ 0] := 1;
          FOutCache[ 1] := AntiRunAway( FInCache[ 10]);
        end;
        else begin
          FOutCache[ 0] := AntiRunAway( FValue * Gain);
          FOutCache[ 1] := AntiRunAway( FInCache[ 10] * FValue * Gain); // sigout := sigin * out
        end;
      end;
    end;

    procedure   TModEnvADSR.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache [ 0]], AppLocale));
    end;

    procedure   TModEnvADSR.GatherData( const aPrefix: string; const aCallback: TDataHandler); // override;
    var
      ATime  : TSignal;
      DTime  : TSignal;
      STime  : TSignal;
      RTime  : TSignal;
      TTime  : TSignal;
      SLevel : TSignal;
      i      : Integer;
      p      : Integer;
      AShape : Integer;
      DShape : Integer;
      RShape : Integer;
      R      : TSignal;
    begin
      ATime  := FInCache[ 5];
      DTime  := FInCache[ 6];
      SLevel := FInCache[ 7];
      RTime  := FInCache[ 8];
      STime  := ( ATime + DTime + RTime) / 3;
      TTime  := ATime + DTime + STime + RTime;
      ATime  := ATime * 67 / TTime;
      DTime  := DTime * 67 / TTime;
      RTime  := RTime * 67 / TTime;
      STime  := STime * 67 / TTime;
      AShape := SignalToInt  ( FInCache[ 2]);
      DShape := SignalToInt  ( FInCache[ 3]);
      RShape := SignalToInt  ( FInCache[ 4]);

      i := 0;
      while i < ATime
      do begin
        case AShape of
          0  : R := LogAttack( i, ATime, 0, 1);
          1  : R := LinAttack( i, ATime, 0, 1);
          else R := ExpAttack( i, ATime, 0, 1);
        end;
        FData[ i] := R;
        Inc( i);
      end;

      p := 0;
      while i < ATime + DTime
      do begin
        case DShape of
          0  : R := LogDecay( p, DTime, 1, SLevel);
          1  : R := LinDecay( p, DTime, 1, SLevel);
          else R := ExpDecay( p, DTime, 1, SLevel);
        end;
        FData[ i] := R;
        Inc( i);
        Inc( p);
      end;

      while i < ATime + DTime + STime
      do begin
        FData[ i] := SLevel;
        Inc( i);
      end;

      p := 0;
      while i < 67
      do begin
        case RShape of
          0  : R := LogDecay( p, RTime, SLevel, 0);
          1  : R := LinDecay( p, RTime, SLevel, 0);
          else R := ExpDecay( p, RTime, SLevel, 0);
        end;
        FData[ i] := R;
        Inc( i);
        Inc( p);
      end;

      aCallback( Self, Format( '%s/%s/viewer', [ aPrefix, Name], AppLocale), FData);
    end;

    function    TModEnvADSR.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
         0 : Result := 'trig';
         1 : Result := 'mode';
         2 : Result := 'ashape';
         3 : Result := 'dshape';
         4 : Result := 'rshape';
         5 : Result := 'attack';
         6 : Result := 'decay';
         7 : Result := 'sustain';
         8 : Result := 'release';
         9 : Result := 'gain';
        10 : Result := 'sigin';
        11 : Result := 'mute';
        12 : Result := 'speed';
        else Result := NO_INPUT;
      end;
    end;

    function    TModEnvADSR.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        1 :  Result := 'sigout';
        else Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModSequencer = class( TMod)
  private
    FScroll  : Integer;
    FSteps   : Integer;
    FCounter : Integer;
    FInState : TSeqInState;
    FOldRes  : Boolean;
  public
}

    constructor TModSequencer.Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := False;
    end;

    procedure   TModSequencer.SetDefaults; // override;
    begin
      inherited;
      FInCache[ 2] := 1; // Make seq run forward by default.
    end;

    function    TModSequencer.Category: string; // override;
    begin
      Result := 'sequencer';
    end;

    procedure   TModSequencer.SetInternal( const aName: string; aValue: TSignal); // override;
    // todo: work
    var
      // OldCount: Integer;
      NewCount: Integer;
    begin
      if SameText( aName, 'steps')
      then begin
        // OldCount := Ceil( GetInternal( 'steps'));
        NewCount := Max( 1, Ceil( aValue));
        if NewCount > InputCount - 3
        then NewCount := InputCount - 3;
        FSteps := NewCount;
      end;
    end;

    function    TModSequencer.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'steps')
      then Result := InputCount - 3
      else if SameText( aName, 'scroll')
      then Result := FScroll
      else Result := inherited;
    end;

    procedure   TModSequencer.DoSlowTick; // override;
    begin
      if not FOldRes and SignalToLogic( FInCache[ 1])
      then begin
        FInState := seqIdle;
        FCounter := 0;
      end;

      case FInState of

        seqIdle :
          begin
            if SignalToLogic( FInCache[ 0])
            then begin
              FInState := seqHigh;
              if SignalToLogic( FInCache[ 2])
              then FCounter := ( FCounter + 1) mod FSteps     // ( InputCount - 3)
              else begin
                if FCounter <= 0
                then FCounter := FSteps - 1 // InputCount - 4
                else Dec( FCounter);
              end;
              FOutCache[ 0] := FInCache[ FCounter + 3] - 69 / 128;
            end;
          end;

        seqHigh :
          begin
            if FInCache[ 0] <= 0
            then FInState := seqIdle;
          end;

      end;
      FOldRes := SignalToLogic( FInCache[ 1]);
    end;

    procedure   TModSequencer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    var
      i : Integer;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, FInCache [ 0]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f', [ aPrefix, Name, FInCache [ 1]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'  , [ aPrefix, Name, FInCache [ 2]], AppLocale));
      for i := 0 to InputCount - 4
      do begin
        if FCounter = i
        then aCallback( Self, Format( '%s/%s/step%d/1.0', [ aPrefix, Name, i + 1], AppLocale))
        else aCallback( Self, Format( '%s/%s/step%d/0.0', [ aPrefix, Name, i + 1], AppLocale));
      end;
    end;

    function    TModSequencer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 : Result := 'trig';
        1 : Result := 'reset';
        2 : Result := 'dir';
        else begin
          if ( anIndex >= 0) and ( anIndex < InputCount)
          then Result := Format( 'step%d', [ anIndex - 2], AppLocale)
          else Result := NO_INPUT;;
        end;
      end;
    end;

    function    TModSequencer.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;

    function    TModSequencer.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'steps';
        1 :  Result := 'scroll';
        else Result := NO_INTERNAL;
      end;
    end;

{ ========
  TModTextSequencer = class( TMod)
  private
    FCounter : Integer;
    FInState : TSeqInState;
    FOldRes  : Boolean;
  public
}

    constructor TModTextSequencer.Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := False;
    end;

    procedure   TModTextSequencer.SetDefaults; // override;
    begin
      inherited;
      FInCache[ 2] := 1; // Make seq run forward by default.
    end;

    function    TModTextSequencer.Category: string; // override;
    begin
      Result := 'sequencer';
    end;

    procedure   TModTextSequencer.SetInternal( const aName: string; aValue: TSignal); // override;
    begin
    end;

    function    TModTextSequencer.GetInternal( const aName: string): TSignal; // override;
    begin
      Result := inherited;
    end;

    procedure   TModTextSequencer.DoSlowTick; // override;
    begin
      if not FOldRes and SignalToLogic( FInCache[ 1])
      then begin
        FInState := seqIdle;
        FCounter := 0;
      end;

      case FInState of

        seqIdle :
          begin
            if SignalToLogic( FInCache[ 0])
            then begin
              FInState := seqHigh;
              if FInCache[ 3] = 0
              then begin
                if SignalToLogic( FInCache[ 2])
                then FCounter := ( FCounter + 1) mod ( InputCount - 4)
                else begin
                  if FCounter <= 0
                  then FCounter := InputCount - 5
                  else Dec( FCounter);
                end;
              end
              else FCounter := Random( InputCount - 4);
              FOutCache[ 0] := FInCache[ FCounter + 4] - 69 / 128;
            end;
          end;

        seqHigh :
          begin
            if FInCache[ 0] <= 0
            then FInState := seqIdle;
          end;

      end;
      FOldRes := SignalToLogic( FInCache[ 1]);
    end;

    procedure   TModTextSequencer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f' , [ aPrefix, Name, FInCache [ 0] ], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f', [ aPrefix, Name, FInCache [ 1] ], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'  , [ aPrefix, Name, FInCache [ 2] ], AppLocale));
      aCallback( Self, Format( '%s/%s/step/%d' , [ aPrefix, Name, FCounter   + 1], AppLocale));
      aCallback( Self, Format( '%s/%s/steps/%d', [ aPrefix, Name, InputCount - 4], AppLocale));
    end;

    function    TModTextSequencer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 : Result := 'trig';
        1 : Result := 'reset';
        2 : Result := 'dir';
        3 : Result := 'mode';
        else begin
          if ( anIndex >= 0) and ( anIndex < InputCount)
          then Result := Format( 'step%d', [ anIndex - 3], AppLocale)
          else Result := NO_INPUT;;
        end;
      end;
    end;

    function    TModTextSequencer.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;

    function    TModTextSequencer.InternalName( anIndex: Integer): string; //override;
    begin
      Result := inherited;
    end;

    procedure   TModTextSequencer.SetStepValues( const aValues: TSignals);
    begin
      while not GrabTickLock do;
      try
        SetInputs( 4, aValues);
      finally
        ReleaseTickLock;
      end;
    end;


{ ========
  TModSeqStep = class( TMod)
  private
    FTime    : TSignal;
    FState   : TSeqStepState;
    FOldRes  : Boolean;
  public
}

    constructor TModSeqStep.Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := False;
    end;

    procedure   TModSeqStep.SetDefaults; // override;
    begin
      inherited;
      FInCache[ 4] := 1; // Make seq run forward by default.
    end;

    function    TModSeqStep.Category: string; // override;
    begin
      Result := 'sequencer';
    end;

    procedure   TModSeqStep.DoSlowTick; // override;
    begin
      if not FOldRes and SignalToLogic( FInCache[ 3])
      then begin                                                       // Reset from inactive to active
        FState        := sssIdle;
        FOutCache[ 0] := LogicToSignal( False);                        // Do not trigger next in forward  direction
        FOutCache[ 1] := LogicToSignal( False);                        // Do not trigger next in backward direction
      end;

      case FState of

        sssIdle :
          begin
            if SignalToLogic( FInCache[ 4])
            then begin                                                 // running Forwards
              if SignalToLogic( FInCache[ 0])                          // Forward trig in
              then begin
                FState        := sssActive;
                FTime         := SlowTimeToSampleCount( FInCache[ 6]); // Duration
                FOutCache[ 0] := LogicToSignal( False);                // Do not trigger next in forward  direction
                FOutCache[ 1] := LogicToSignal( False);                // Do not trigger next in backward direction
              end;
            end
            else begin                                                 // running Backwards
              if SignalToLogic( FInCache[ 1])                          // Backward trig in
              then begin
                FState        := sssActive;
                FTime         := SlowTimeToSampleCount( FInCache[ 6]); // Duration
                FOutCache[ 0] := LogicToSignal( False);                // Do not trigger next in forward  direction
                FOutCache[ 1] := LogicToSignal( False);                // Do not trigger next in backward direction
              end;
            end;
          end;

        sssActive :
          begin
            FTime := FTime - 1;
            if FTime <= 0
            then begin
              FState := sssWaiting;
              if SignalToLogic( FInCache[ 4])                          // running Forwards
              then FOutCache[ 0] := LogicToSignal( True)               // Trigger next in forward  direction
              else FOutCache[ 1] := LogicToSignal( True);              // Trigger next in backward direction
            end;
          end;

        sssWaiting :
          begin
            if SignalToLogic( FInCache[ 4])
            then begin                                                 // running Forwards
              if not SignalToLogic( FInCache[ 0])                      // Forward trig in fell off
              then begin
                FState := sssIdle;
              end;
            end
            else begin                                                 // running Backwards
              if not SignalToLogic( FInCache[ 1])                      // Backward trig in fell off
              then begin
                FState := sssIdle;
              end;
            end;
          end;

      end;

      if FState = sssActive
      then FOutCache[ 2] := FInCache[ 5] - 69 / 128                    // Own value to output
      else FOutCache[ 2] := FInCache[ 2];                              // Chained value to output (was compnsated there already)

      FOldRes := SignalToLogic( FInCache[ 3]);
    end;

    procedure   TModSeqStep.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trigfwd/%f'   , [ aPrefix, Name, FInCache [ 0]], AppLocale));
      aCallback( Self, Format( '%s/%s/trigbwd/%f'   , [ aPrefix, Name, FInCache [ 1]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f'     , [ aPrefix, Name, FInCache [ 3]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'       , [ aPrefix, Name, FInCache [ 4]], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f'    , [ aPrefix, Name, IfThen( FState = sssActive, 1.0, 0.0)], AppLocale));
    end;

    function    TModSeqStep.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'triginfwd';
        1 :  Result := 'triginbwd';
        2 :  Result := 'chain';
        3 :  Result := 'reset';
        4 :  Result := 'dir';
        5 :  Result := 'note';
        6 :  Result := 'time';
        else Result := NO_INPUT;
      end;
    end;

    function    TModSeqStep.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'trigoutfwd';
        1 :  Result := 'trigoutbwd';
        2 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;

{ ========
  TModSeqClockStep = class( TMod)
  private
    FCount      : Integer;
    FState      : TSeqStepState;
    FOldRes     : Boolean;
    FOldClock   : Boolean;
    FTrigInFwd  : Boolean;
    FTrigInBwd  : Boolean;
    FDir        : Boolean;
    FTrigOutFwd : Boolean;
    FTrigOutBwd : Boolean;
  public
}

    constructor TModSeqClockStep.Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := False;
    end;

    procedure   TModSeqClockStep.SetDefaults; // override;
    begin
      inherited;
      FInCache[ 4] := 1; // Make seq run forward by default.
    end;

    function    TModSeqClockStep.Category: string; // override;
    begin
      Result := 'sequencer';
    end;

    procedure   TModSeqClockStep.DoSlowTick; // override;
    begin

      if not FOldRes and SignalToLogic( FInCache[ 3])
      then begin                                                       // Reset from inactive to active - reset is asynchroneous
        FState        := sssIdle;
        FTrigOutFwd   := False;
        FTrigOutBwd   := False;
        FOutCache[ 0] := LogicToSignal( False);                        // Do not trigger next in forward  direction
        FOutCache[ 1] := LogicToSignal( False);                        // Do not trigger next in backward direction
        FOutCache[ 2] := FInCache[ 2];                                 // Chained value to output
      end;

      FTrigInFwd  := SignalToLogic( FInCache[ 0]);
      FTrigInBwd  := SignalToLogic( FInCache[ 1]);
      FDir        := SignalToLogic( FInCache[ 4]);

      case FState of

        sssIdle :
          begin
            if ( FDir and FTrigInFwd) or (( not FDir) and FTrigInBwd)
            then begin                                               // Trigger seen
              FState      := sssActive;
              FCount      := Ceil( FInCache[ 7]);                    // Duration
              FTrigOutFwd := False;                                  // Do not trigger next in forward  direction
              FTrigOutBwd := False;                                  // Do not trigger next in backward direction
            end;
          end;

        sssWaiting :
          begin
            if ( FDir and ( not FTrigInFwd)) or (( not FDir) and ( not FTrigInBwd))
            then begin
              FState      := sssIdle;                                // Trigger fell off - go idle
              FTrigOutFwd := False;                                  // Do not trigger next in forward  direction
              FTrigOutBwd := False;                                  // Do not trigger next in backward direction
            end;
          end;
      end;

      if not FOldClock and SignalToLogic( FInCache[ 5])
      then begin                                                       // Clock from inactive to active - state changes should only occur here

        case FState of
          sssActive :
            begin
              FCount := FCount - 1;
              if FCount <= 1
              then begin
                FCount := 0;
                FState := sssWaiting;
                if FDir                                                // running Forwards
                then begin
                  FTrigOutFwd := True;                                 // Trigger next in forward  direction
                  FTrigOutBwd := False;                                // No Trigger next in backward  direction
                end
                else begin
                  FTrigOutFwd := False;                                // No Trigger next in forward direction
                  FTrigOutBwd := True;                                 // Trigger next in backward direction
                end;
              end;
            end;
        end;

      end;

      FOutCache[ 0] := LogicToSignal( FTrigOutFwd);                  // Clock out internal state to ouptputs
      FOutCache[ 1] := LogicToSignal( FTrigOutBwd);
      if FState = sssActive
      then FOutCache[ 2] := FInCache[ 6] - 69 / 128                  // Own     value to output
      else FOutCache[ 2] := FInCache[ 2];                            // Chained value to output (was compensated there already)

      FOldRes   := SignalToLogic( FInCache[ 1]);
      FOldClock := SignalToLogic( FInCache[ 5]);
    end;


    procedure   TModSeqClockStep.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trigfwd/%f', [ aPrefix, Name, FInCache [ 0]], AppLocale));
      aCallback( Self, Format( '%s/%s/trigbwd/%f', [ aPrefix, Name, FInCache [ 1]], AppLocale));
      aCallback( Self, Format( '%s/%s/reset/%f'  , [ aPrefix, Name, FInCache [ 3]], AppLocale));
      aCallback( Self, Format( '%s/%s/dir/%f'    , [ aPrefix, Name, FInCache [ 4]], AppLocale));
      aCallback( Self, Format( '%s/%s/clock/%f'  , [ aPrefix, Name, FInCache [ 5]], AppLocale));
      aCallback( Self, Format( '%s/%s/count/%d'  , [ aPrefix, Name, FCount], AppLocale));
      aCallback( Self, Format( '%s/%s/active/%f' , [ aPrefix, Name, IfThen( FState = sssActive, 1.0, 0.0)], AppLocale));
    end;

    function    TModSeqClockStep.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'triginfwd';
        1 :  Result := 'triginbwd';
        2 :  Result := 'chain';
        3 :  Result := 'reset';
        4 :  Result := 'dir';
        5 :  Result := 'clock';
        6 :  Result := 'note';
        7 :  Result := 'counts';
        else Result := NO_INPUT;
      end;
    end;

    function    TModSeqClockStep.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'trigoutfwd';
        1 :  Result := 'trigoutbwd';
        2 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDelay = class( TMod)
  private
    FLength    : TSignal;
    FMemory    : TSignals;
    FInPointer : Integer;
  public
}

    procedure   TModDelay.SetDefaults; // override;
    begin
      inherited;
      FInCache[ 1] := 1.0; // Set maximum delay time if nothing connected
    end;

    function    TModDelay.Category: string; // override;
    begin
      Result := 'delay';
    end;

    destructor  TModDelay.Destroy; // override;
    begin
      inherited;
      FMemory.Free;
    end;

    procedure   TModDelay.SetInternal( const aName: string; aValue: TSignal); // override;
    var
      aMemory: TSignals;
      bMemory: TSignals;
    begin
      if SameText( aName, 'length')
      then begin
        if aValue < 0.00001
        then aValue := 0.00001;
        if aValue <> FDlyTime
        then begin
          aMemory  := TSignals.Create( Ceil( TimeToSampleCount( aValue)));
          bMemory  := FMemory;

          while not GrabTickLock do;
          try
            FMemory  := aMemory;
            FDlyTime := aValue;
          finally
            ReleaseTickLock;
          end;

          FreeAndNil( bMemory);
        end;
      end
      else inherited;
    end;

    function    TModDelay.GetInternal( const aName: string): TSignal; // override;
    begin
      if SameText( aName, 'length')
      then Result := FDlyTime
      else Result := inherited;
    end;

    procedure   TModDelay.DoTick; // override;
    var
      OutPointerVar : TSignal;
      SampCount     : Integer;
    begin
      if GrabTickLock
      then begin
        try
          SampCount     := FMemory.Count;
          FInPointer    := FInPointer mod SampCount;
          FOutCache[ 1] := FMemory.Signal[ FInPointer];
          FMemory.Signal[ FInPointer] := Clip( FInCache[ 0], -16, 16);
          OutPointerVar := FModulo( SampCount + FInPointer - Clip( FInCache[ 1], 0.0, 1.0) * ( TimeToSampleCount( FDlyTime) - 1), SampCount);
          FOutCache[ 0] := AntiRunAway( FMemory.Interpolate[ OutPointerVar]);
          FInPointer    := ( FInPointer + 1) mod SampCount;
        finally
          ReleaseTickLock;
        end;
      end;
    end;

    function    TModDelay.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'in';
        1 :  Result := 'delay';
        else Result := NO_INPUT;
      end;
    end;

    function    TModDelay.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        1 :  Result := 'outfixed';
        else Result := NO_OUTPUT;
      end;
    end;

    function    TModDelay.InternalName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'length';
        else Result := NO_INTERNAL;
      end;
    end;

{ ========
  TModWave = class( TMod)
  private
    FFileName    : string;
    FMemory      : TSignals;
    FDuration    : string;
  public
    property    FileName: string read FFileName write SetFileName;
  private
}

    procedure   TModWave.SetDuration( aValue: Cardinal);
    begin
      if aValue < 1000
      then FDuration := Format( '%.3f ms', [ aValue / 1.0], AppLocale)
      else FDuration := Format( '%.3f s', [ aValue / 1000.0], AppLocale);
    end;

    procedure   TModWave.SwapMemoryFor( aMemory: TSignalArray);
    var
      aNewMemory  : TSignals;
      anOldMemory : TSignals;
      i           : Integer;
    begin
      aNewMemory  := TSignals.Create( Length( aMemory));
      for i := 0 to Length( aMemory) - 1
      do aNewMemory.Signal[ i] := aMemory[ i];

      while not GrabTickLock do;
      try
        anOldMemory := FMemory;
      finally
        ReleaseTickLock;
      end;

      FMemory     := aNewMemory;
      FreeAndNil( anOldMemory);
    end;

    procedure   TModWave.LoadFile;
    const
      BufSize = 8192;
    type
      TSample = SmallInt;
      TBuffer = array[ 0 .. BufSize - 1] of TSample;

      procedure SetSamplesStereo16bit44100Hz( var aMemory: TSignalArray; aBuf: TBuffer; aSize: Cardinal);
      const
        MaxVal = High( TSample) / 2.0;
      var
        i : Cardinal;
        p : Cardinal;
        n : Cardinal;
      begin
        Assert( not Odd( aSize  div 2));
        p := Length( aMemory);
        n := aSize div 4;
        SetLength( aMemory, p + n);
        for i := 0 to n - 1
        do begin
          aMemory[ p] := aBuf[ 2 * i] /  MaxVal + aBuf[ 2 * i + 1] / MAxVal;
          Inc( p);
        end;
      end;

    var
      aMemory   : TSignalArray;
      aWaveFile : TWaveFile;
      aSize     : Cardinal;
      aBuf      : TBuffer;
      aDuration : Cardinal;
      aFormat   : TPCMFormat;
    begin
      SetLength( aMemory, 0);
      aDuration := 0;
      if FileExists( FileName) // todo: and FileSize < 44100 * 2 * 2 * 301 // 44k1, stereo, 16 bit and duration < 301 s.
      then begin
        aWaveFile := TWaveFile.Create( FileName, fmOpenRead);
        try
          if aWaveFile.Valid
          then begin
            aFormat := aWaveFile.PCMFormat;
            if aFormat = Stereo16bit44100Hz
            then begin
              aDuration := aWaveFile.Length;
              aWaveFile.Position := 0;
              if aWaveFile.BeginRead
              then begin
                try
                  while Cardinal( aWaveFile.Position) < aWaveFile.DataSize
                  do begin
                    aSize := aWaveFile.Read( aBuf, BufSize * SizeOf( TSample));
                    case aFormat of
                      Stereo16bit44100Hz : begin SetSamplesStereo16bit44100Hz( aMemory, aBuf, aSize); aDuration := aDuration * 1; end;
                    end;
                  end;
                finally
                  aWaveFile.EndRead;
                end;
              end;
            end;
          end;
        finally
          aWaveFile.Free;
        end;
      end;
      SwapMemoryFor( aMemory);
      SetDuration( aDuration);
    end;

    procedure   TModWave.SetFileName( const aValue: string);
    begin
      if  aValue <> FFileName
      then begin
        FFileName := aValue;
        LoadFile;
      end;
    end;

//  public

    destructor  TModWave.Destroy; // override;
    begin
      inherited;
      FMemory.Free;
    end;

    function    TModWave.Category: string; // override;
    begin
      Result := 'delay';
    end;

    procedure   TModWave.SetStringValue( const aName, aValue: string); // override;
    begin
      if SameText( aName, 'filename')
      then FileName := aValue
      else inherited;
    end;

    function    TModWave.GetStringValue( const aName: string): string; // override;
    begin
      if SameText( aName, 'filename')
      then Result := FFileName
      else Result := inherited;
    end;

    procedure   TModWave.GatherInfo( const aPrefix: string; const aCallback: TInfoHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/duration/%s', [ aPrefix, Name, FDuration], AppLocale));
    end;

    function    TModWave.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;

    function    TModWave.StringName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'filename';
        else Result := NO_STRING;
      end;
    end;



{ ========
  TModWavePlayer = class( TModWave)
  private
    FPlayPointer : TSignal;
    FRunning     : Boolean;
  public
}

    procedure   TModWavePlayer.SetDefaults; // override;
    begin
      inherited;
      FInCache[ 0] := 0.0; // Set zero speed
      FInCache[ 1] := 0.0; // Set start position
      FInCache[ 2] := 0.0; // Reset to zero
      SetDuration( 0);
    end;

    procedure   TModWavePlayer.DoTick; // override;
    var
      SampCount : Integer;
      Increment : TSignal;
      Position  : TSignal;
    begin
      if ( not FOldRes) and SignalToLogic( FInCache[ 2])
      then FPlayPointer := 0.0;
      if GrabTickLock
      then begin
        try
          if Assigned( FMemory)
          then begin
            SampCount     := FMemory.Count;
            Increment     := Clip( FInCache[ 0], - 16, 16);
            Position      := ( 1 + Clip( FInCache[ 1],  0, 1)) * SampCount;
            FPlayPointer  := FModulo( FPlayPointer + Increment + SampCount, SampCount);
            Position      := FModulo( Position + FPlayPointer, SampCount);
            FOutCache[ 0] := FMemory.Interpolate[ Position];
          end;
        finally
          ReleaseTickLock;
        end;
      end;
      FOldRes := SignalToLogic( FInCache[ 2]);
    end;

    procedure   TModWavePlayer.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/res/%f', [ aPrefix, Name, FInCache[ 2]], AppLocale));
    end;

    function    TModWavePlayer.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'speed';
        1 :  Result := 'position';
        2 :  Result := 'res';
        else Result := NO_INPUT;
      end;
    end;



{ ========
  TModSampleAndHold = class( TMod)
  private
    FLastTrig: Boolean;
  public
}

    function    TModSampleAndHold.Category: string; // override;
    begin
      Result := 'switch';
    end;

    procedure   TModSampleAndHold.DoTick; // override;
    var
      Trig : Boolean;
    begin
      Trig := SignalToLogic( FInCache[ 0]);
      if not FLastTrig and Trig
      then FOutCache[ 0] := FInCache[ 1];
      FLastTrig := Trig;
    end;

    procedure   TModSampleAndHold.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache[ 0]], AppLocale));
    end;

    function    TModSampleAndHold.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'trig';
        1 :  Result := 'in';
        else Result := NO_INPUT;
      end;
    end;

    function    TModSampleAndHold.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModTrackAndHold = class( TMod)
  public
}

    function    TModTrackAndHold.Category: string; // override;
    begin
      Result := 'switch';
    end;

    procedure   TModTrackAndHold.DoTick; // override;
    begin
      if SignalToLogic( FInCache[ 0])
      then FOutCache[ 0] := FInCache[ 1];
    end;

    procedure   TModTrackAndHold.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/trig/%f', [ aPrefix, Name, FInCache[ 0]], AppLocale));
    end;

    function    TModTrackAndHold.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'trig';
        1 :  Result := 'in';
        else Result := NO_INPUT;
      end;
    end;

    function    TModTrackAndHold.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModSwitch2to1 = class( TMod)
  public
}

    function    TModSwitch2to1.Category: string; // override;
    begin
      Result := 'switch';
    end;

    procedure   TModSwitch2to1.DoTick; // override;
    begin
      if SignalToLogic( FInCache[ 0])
      then FOutCache[ 0] := FInCache[ 2]
      else FOutCache[ 0] := FInCache[ 1];
    end;

    procedure   TModSwitch2to1.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    begin
      aCallback( Self, Format( '%s/%s/control/%f', [ aPrefix, Name, FInCache[ 1]], AppLocale));
    end;

    function    TModSwitch2to1.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'control';
        1 :  Result := 'in1';
        2 :  Result := 'in2';
        else Result := NO_INPUT;
      end;
    end;

    function    TModSwitch2to1.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModXFade = class( TMod)
  public
}

    function    TModXFade.Category: string; // override;
    begin
      Result := 'mixer';
    end;

    procedure   TModXFade.DoTick; // override;
    var
      Control: TSignal;
    begin
      Control    := FInCache[ 0];
      FOutCache[ 0] := AntiRunAway((( 1 - Control) * FInCache[ 1] + ( 1 + Control) * FInCache[ 2]) / 2);
    end;

    function    TModXFade.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'control';
        1 :  Result := 'in1';
        2 :  Result := 'in2';
        else Result := NO_INPUT;
      end;
    end;

    function    TModXFade.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;



{ ========
  TModPan = class( TMod)
  public
}

    function    TModPan.Category: string; // override;
    begin
      Result := 'mixer';
    end;

    procedure   TModPan.DoTick; // override;
    var
      Control: TSignal;
      Inp    : TSignal;
    begin
      Control    := FInCache[ 0];
      Inp        := FInCache[ 1];
      FOutCache[ 0] := AntiRunAway((( 1 - Control) * Inp) / 2);
      FOutCache[ 1] := AntiRunAway((( 1 + Control) * Inp) / 2);
    end;

    function    TModPan.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'control';
        1 :  Result := 'in';
        else Result := NO_INPUT;
      end;
    end;

    function    TModPan.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out1';
        1 :  Result := 'out2';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModQuantize = class( TMod)
  public
}

    function    TModQuantize.Category: string; // override;
    begin
      Result := 'signal conditioning';
    end;

    procedure   TModQuantize.DoTick; // override;
    var
      Steps: TSignal;
    begin
      Steps := Clip( 1000 * FInCache[ 0], 1, 1000);
      if SignalToLogic( FInCache[ 2])
      then FOutCache[ 0] := 0
      else FOutCache[ 0] := Round( Steps * FInCache[ 1]) / Steps;
    end;

    function    TModQuantize.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'qsteps';
        1 :  Result := 'in';
        2 :  Result := 'mute';
        else Result := NO_INPUT;
      end;
    end;

    function    TModQuantize.OutputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'out';
        else Result := NO_OUTPUT;
      end;
    end;


{ ========
  TModDisplay = class( TMod)
  private
    FMax1: TSignal;
    FMax2: TSignal;
  public
}

    constructor TModDisplay.Create( aParent: TSynthPatch; const aName: string; anInputCount: Integer); // override;
    begin
      inherited;
      FIsSlow := True;
      FIsFast := False;
    end;

    function    TModDisplay.Category: string; // override;
    begin
      Result := 'utility';
    end;

    procedure   TModDisplay.DoSlowTick; // override;
    begin
      FMax1 := Max( FMax1, Abs( FInCache[ 0]));
      FMax2 := Max( FMax2, Abs( FInCache[ 1]));
    end;

    procedure   TModDisplay.GatherLights( const aPrefix: string; const aCallback: TLightsHandler); // override;
    const
      LOWVAL  = -45.0;
      HIGHVAL =  18.0;
      LOWLED  =   0.0;
      HIGHLED =   1.0;
    var
      adB1 : TSignal;
      adB2 : TSignal;
    begin
      adB1 := RangeMap( Clip( SignalTodB( FMax1), LOWVAL, HIGHVAL), LOWVAL, HIGHVAL, LOWLED, HIGHLED);
      adB2 := RangeMap( Clip( SignalTodB( FMax2), LOWVAL, HIGHVAL), LOWVAL, HIGHVAL, LOWLED, HIGHLED);
      aCallback( Self, Format( '%s/%s/level1/%f', [ aPrefix, Name, adB1], AppLocale));
      aCallback( Self, Format( '%s/%s/level2/%f', [ aPrefix, Name, adB2], AppLocale));
      FMax1 := 0;
      FMax2 := 0;
    end;

    function    TModDisplay.InputName( anIndex: Integer): string; // override;
    begin
      case anIndex of
        0 :  Result := 'in1';
        1 :  Result := 'in2';
        else Result := NO_INPUT;
      end;
    end;

    function    TModDisplay.OutputName( anIndex: Integer): string; // override;
    begin
      Result := NO_OUTPUT;
    end;


initialization

  InitializeUnit;

finalization

  FinalizeUnit;

end.
